// Copyright 2016-2021 JetBrains s.r.o.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package com.goide.psi.impl.codeFragment;

import com.goide.GoCodeInsightFixtureTestCase;
import com.goide.psi.*;
import com.goide.psi.impl.GoElementFactory;
import com.intellij.openapi.command.WriteCommandAction;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.io.FileUtil;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.psi.PsiElement;
import com.intellij.psi.util.PsiTreeUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.File;
import java.io.IOException;
import java.nio.file.Paths;
import java.util.Objects;

public abstract class GoCodeFragmentTestBase extends GoCodeInsightFixtureTestCase {
  @NotNull
  protected GoCodeFragment openFragmentInEditor() {
    String contextText = Objects.requireNonNull(readTestFile("context.go"));
    String fragmentText = Objects.requireNonNull(readTestFile("fragment.go"));
    String fragmentImportsText = readTestFile("fragmentImports.go");
    return openFragmentInEditor(contextText, fragmentText, fragmentImportsText);
  }

  @NotNull
  protected GoCodeFragment openFragmentInEditor(@NotNull String contextText,
                                                @NotNull String fragmentText,
                                                @Nullable String fragmentImportsText) {
    return openFragmentInEditor(contextText, fragmentText, fragmentImportsText, getCodeFragmentProvider());
  }

  @NotNull
  abstract protected CodeFragmentProvider getCodeFragmentProvider();

  @NotNull
  protected GoCodeFragment openFragmentInEditor(@NotNull String contextText,
                                                @NotNull String fragmentText,
                                                @Nullable String fragmentImportsText,
                                                @NotNull CodeFragmentProvider provider) {
    myFixture.configureByText("context.go", contextText);
    PsiElement context = findElementAtCaretOrInSelection();
    GoCodeFragment fragment = provider.create(getProject(), StringUtil.replace(fragmentText, "<caret>", ""), true, context, true);
    if (fragmentImportsText != null) {
      GoFile fileWithFragmentImports = GoElementFactory.createFileFromText(getProject(), fragmentImportsText);
      GoImportList importList = PsiTreeUtil.findChildOfType(fileWithFragmentImports, GoImportList.class);
      assertNotNull("Import list for fragment is incorrect, it must be a valid go file", importList);
      WriteCommandAction.runWriteCommandAction(getProject(), () -> {
        for (GoImportDeclaration declaration : importList.getImportDeclarationList()) {
          for (GoImportSpec spec : declaration.getImportSpecList()) {
            fragment.addImport(spec.getPath(), spec.getAlias());
          }
        }
      });
    }

    myFixture.openFileInEditor(fragment.getVirtualFile());
    int caretOffset = fragmentText.indexOf("<caret>");
    if (caretOffset > -1) {
      myFixture.getEditor().getCaretModel().moveToOffset(caretOffset);
    }
    return fragment;
  }

  @Nullable
  private String readTestFile(String s) {
    File file = Paths.get(getTestDataPath(), getTestName(true), s).toFile();
    try {
      return file.exists() ? FileUtil.loadFile(file, true) : null;
    }
    catch (IOException e) {
      fail("failed to read test file: " + e.getMessage());
      return null;
    }
  }

  @FunctionalInterface
  protected interface CodeFragmentProvider {
    GoCodeFragment create(@NotNull Project project, @NotNull CharSequence text, boolean isPhysical, @Nullable PsiElement context,
                          boolean supportsImports);
  }
}
